@@ -16,11 +16,14 @@
 //! Command-line interface for the EndBASIC language.
 
 // Keep these in sync with other top-level files.
+#![allow(clippy::await_holding_refcell_ref)]
 #![warn(anonymous_parameters, bad_style, missing_docs)]
 #![warn(unused, unused_extern_crates, unused_import_braces, unused_qualifications)]
 #![warn(unsafe_code)]
 
-use anyhow::{anyhow, Error, Result};
+use anyhow::{anyhow, Result};
+use endbasic_std::store::{FileStore, Store};
+use endbasic_std::terminal::TerminalConsole;
 use futures_lite::future::block_on;
 use getopts::Options;
 use std::cell::RefCell;
@@ -45,14 +48,6 @@ impl UsageError {
     }
 }
 
-/// Flattens all causes of an error into a single string.
-fn flatten_causes(err: &Error) -> String {
-    err.chain().fold(String::new(), |flattened, cause| {
-        let flattened = if flattened.is_empty() { flattened } else { flattened + ": " };
-        flattened + &format!("{}", cause)
-    })
-}
-
 /// Consumes and returns the program name from `env::Args`.
 ///
 /// If the program name cannot be obtained, return `default_name` instead.
@@ -82,7 +77,7 @@ fn help(name: &str, opts: &Options) -> Result<i32> {
 /// Prints version information following the GNU Standards format.
 fn version() -> Result<i32> {
     println!("EndBASIC {}", env!("CARGO_PKG_VERSION"));
-    println!("Copyright 2020 Julio Merino");
+    println!("Copyright 2020-2021 Julio Merino");
     println!("License Apache Version 2.0 <http://www.apache.org/licenses/LICENSE-2.0>");
     Ok(0)
 }
@@ -109,34 +104,46 @@ fn get_programs_dir(flag: Option<String>) -> Result<PathBuf> {
     Ok(dir.unwrap())
 }
 
-/// Executes the `path` program in a fresh machine.
-fn run<P: AsRef<Path>>(path: P) -> endbasic_core::exec::Result<i32> {
-    let console = Rc::from(RefCell::from(endbasic::TextConsole::from_stdio()?));
-    let exit_code = Rc::from(RefCell::from(None));
-    let mut machine = {
-        let mut builder = endbasic_core::exec::MachineBuilder::default()
-            .add_command(endbasic_core::repl::ExitCommand::new(exit_code.clone()))
-            .add_commands(endbasic_core::console::all_commands(console))
-            .add_functions(endbasic_core::strings::all_functions());
-        builder = endbasic_core::numerics::add_all(builder);
-        builder.build()
-    };
-
-    let mut input = File::open(path)?;
-    match block_on(machine.exec(&mut input)) {
-        Ok(()) => (),
-        Err(e) => {
-            if exit_code.borrow().is_some() {
-                if let endbasic_core::exec::Error::IoError(e) = e {
-                    debug_assert!(e.kind() == io::ErrorKind::UnexpectedEof);
-                }
-            } else {
-                return Err(e);
-            }
-        }
+/// Creates a new store backed by `dir` and overlays the built-in demos.
+fn new_store_with_demos(dir: &Path) -> Rc<RefCell<dyn Store>> {
+    if dir == Path::new(":memory:") {
+        Rc::from(RefCell::from(endbasic::demos::DemoStoreOverlay::new(
+            endbasic_std::store::InMemoryStore::default(),
+        )))
+    } else {
+        Rc::from(RefCell::from(endbasic::demos::DemoStoreOverlay::new(FileStore::new(dir))))
     }
-    let exit_code = exit_code.borrow();
-    Ok(exit_code.unwrap_or(0))
+}
+
+/// Enters the interactive interpreter.
+///
+/// `dir` specifies the directory that the interpreter will use for any commands that manipulate
+/// files.  The special name `:memory:` makes the interpreter use an in-memory only store.
+fn run_repl_loop(dir: &Path) -> io::Result<i32> {
+    let console = Rc::from(RefCell::from(TerminalConsole::from_stdio()?));
+    let store = new_store_with_demos(dir);
+    let mut machine = endbasic_std::interactive_machine(console.clone(), store.clone());
+    endbasic::print_welcome(console.clone())?;
+    endbasic::try_load_autoexec(&mut machine, console.clone(), store)?;
+    block_on(endbasic::run_repl_loop(&mut machine, console))
+}
+
+/// Executes the `path` program in a fresh machine.
+fn run_script<P: AsRef<Path>>(path: P) -> endbasic_core::exec::Result<i32> {
+    let console = Rc::from(RefCell::from(TerminalConsole::from_stdio()?));
+    let mut machine = endbasic_std::scripting_machine(console);
+    let mut input = File::open(path)?;
+    Ok(block_on(machine.exec(&mut input))?.as_exit_code())
+}
+
+/// Executes the `path` program in a fresh machine allowing any interactive-only calls.
+///
+/// `dir` has the same meaning as the parameter passed to `run_repl_loop`.
+fn run_interactive<P: AsRef<Path>>(path: P, dir: &Path) -> endbasic_core::exec::Result<i32> {
+    let console = Rc::from(RefCell::from(TerminalConsole::from_stdio()?));
+    let mut machine = endbasic_std::interactive_machine(console, new_store_with_demos(dir));
+    let mut input = File::open(path)?;
+    Ok(block_on(machine.exec(&mut input))?.as_exit_code())
 }
 
 /// Version of `main` that returns errors to the caller for reporting.
@@ -145,6 +152,7 @@ fn safe_main(name: &str, args: env::Args) -> Result<i32> {
 
     let mut opts = Options::new();
     opts.optflag("h", "help", "show command-line usage information and exit");
+    opts.optflag("i", "interactive", "force interactive mode when running a script");
     opts.optopt("", "programs-dir", "directory where user programs are stored", "PATH");
     opts.optflag("", "version", "show version information and exit");
     let matches = opts.parse(args)?;
@@ -160,46 +168,40 @@ fn safe_main(name: &str, args: env::Args) -> Result<i32> {
     match matches.free.as_slice() {
         [] => {
             let programs_dir = get_programs_dir(matches.opt_str("programs-dir"))?;
-            Ok(endbasic::run_repl_loop(&programs_dir)?)
+            Ok(run_repl_loop(&programs_dir)?)
+        }
+        [file] => {
+            if matches.opt_present("interactive") {
+                let programs_dir = get_programs_dir(matches.opt_str("programs-dir"))?;
+                Ok(run_interactive(file, &programs_dir)?)
+            } else {
+                Ok(run_script(file)?)
+            }
         }
-        [file] => Ok(run(file)?),
         [_, ..] => Err(UsageError::new("Too many arguments").into()),
     }
 }
 
 fn main() {
     let (name, args) = program_name(env::args(), "endbasic");
-    match safe_main(&name, args) {
-        Ok(code) => process::exit(code),
+    let exit_code = match safe_main(&name, args) {
+        Ok(code) => code,
         Err(e) => {
             if let Some(e) = e.downcast_ref::<UsageError>() {
                 eprintln!("Usage error: {}", e);
                 eprintln!("Type {} --help for more information", name);
-                process::exit(2);
+                2
             } else if let Some(e) = e.downcast_ref::<getopts::Fail>() {
                 eprintln!("Usage error: {}", e);
                 eprintln!("Type {} --help for more information", name);
-                process::exit(2);
+                2
             } else {
-                eprintln!("{}: {}", name, flatten_causes(&e));
-                process::exit(1);
+                eprintln!("{}: {}", name, e);
+                1
             }
         }
-    }
-}
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-
-    #[test]
-    fn flatten_causes_one() {
-        assert_eq!("the error", flatten_causes(&anyhow!("the error")));
-    }
-
-    #[test]
-    fn flatten_causes_several() {
-        let err = anyhow!("first").context("second").context("and last");
-        assert_eq!("and last: second: first", flatten_causes(&err));
-    }
+    };
+    // There should not be any interesting destructors left behind when calling this, or else they
+    // will not run.
+    process::exit(exit_code);
 }
